@@ -1,5 +1,6 @@
 import abc
 import itertools
+import logging
 import sys
 from collections import defaultdict
 from enum import Enum
@@ -8,13 +9,14 @@ from typing import Union, List, Any, Callable, Type, Dict, Set, TypeVar
 
 from yattag import Doc
 
-from pytconf.color_utils import print_highlight, color_hi, print_title, color_ok, color_warn, print_warn
-from pytconf.convert import convert_string_to_int, convert_int_to_string, \
-    convert_string_to_list_int, \
-    convert_list_int_to_string, convert_string_to_list_str, convert_list_str_to_string, convert_string_to_string, \
-    convert_string_to_bool, convert_bool_to_string, convert_string_to_int_or_none, convert_int_or_none_to_string, \
-    convert_string_to_int_default
+from pytconf.color_utils import print_highlight, color_hi, print_title, color_ok, color_warn, print_error
+from pytconf.convert import convert_str_to_int, convert_int_to_str, convert_str_to_int_default, \
+    convert_str_to_list_int, convert_list_int_to_str, convert_list_str_to_str, \
+    convert_str_to_list_str, convert_str_to_int_or_none, \
+    convert_int_or_none_to_str, convert_str_to_bool, convert_bool_to_str, convert_str_to_str, \
+    convert_str_to_str_or_none, convert_str_or_none_to_str
 from pytconf.enum_subset import EnumSubset
+from pytconf.errors_collector import ErrorsCollector
 from pytconf.extended_enum import str_to_enum_value, enum_type_to_list_str
 
 PARAMS_ATTRIBUTE = "_params"
@@ -22,6 +24,57 @@ DEFAULT_GROUP_NAME: str = "default"
 SPECIAL_COMMANDS = {"help", "help-suggest", "help-all"}
 
 
+class MetaConfig(type):
+    """
+    base class for all configs
+    """
+
+    def __new__(mcs, name, bases, cls_dict):
+        if name != "Config":
+            params_dict = dict()
+            cls_dict[PARAMS_ATTRIBUTE] = params_dict
+            for k, d in cls_dict.items():
+                if isinstance(d, Param):
+                    # assert d.default is not NO_DEFAULT
+                    params_dict[k] = d
+                    cls_dict[k] = d.default
+        return type.__new__(mcs, name, bases, cls_dict)
+
+    def __init__(cls, name, bases, cls_dict):
+        # print(name, bases, cls_dict)
+        # if len(cls.mro()) > 2:
+        #     register_config(cls, name)
+        #     # print("was subclassed by " + name)
+        if name != "Config":
+            # noinspection PyTypeChecker
+            get_pytconf().register_config(cls, name)
+        # print(name, cls_dict)
+        super(MetaConfig, cls).__init__(name, bases, cls_dict)
+
+
+ParamType = TypeVar('ParamType', bound='Param')
+
+
+class Config(metaclass=MetaConfig):
+    """
+        base class for all configs
+    """
+
+    @classmethod
+    def get_attributes(cls: Any) -> List[str]:
+        return getattr(cls, PARAMS_ATTRIBUTE).keys()
+        # return [attr for attr in dir(cls) if not callable(getattr(cls, attr))
+        #        and not attr.startswith("__")]
+
+    @classmethod
+    def get_params(cls: Any) -> Dict[str, ParamType]:
+        return getattr(cls, PARAMS_ATTRIBUTE)
+
+    @classmethod
+    def get_param_by_name(cls: Any, name: str) -> ParamType:
+        return cls.get_params()[name]
+
+
 class HtmlGen(object):
     def __init__(self):
         document, tag, text, line = Doc().ttl()
@@ -36,33 +89,39 @@ class PytconfConf(object):
         self._configs = set()
         self._config_names = set()
         self.main_function = None
-        self.function_name_to_configs: Dict[str, List[Config]] = dict()
-        self.function_name_to_suggest_configs: Dict[str, List[Config]] = dict()
+        self.function_name_to_configs: Dict[str, List[Type[Config]]] = dict()
+        self.function_name_to_suggest_configs: Dict[str, List[Type[Config]]] = dict()
         self.function_name_to_callable: Dict[str, Callable] = dict()
         self.function_group_names: Dict[str, Set[str]] = defaultdict(set)
+        self.allow_free_args: Dict[str, bool] = dict()
         self.function_group_descriptions: Dict[str, str] = dict()
         self.function_group_list = []
+        self.attribute_to_config: Dict[str, Type[Config]] = dict()
+        self.free_args: List[str] = []
+        self.app_name: Union[str, None] = None
 
     def register_main(self, f):
         self.main_function = f
 
-    def register_config(self, cls, name):
+    def register_config(self, config: Type[Config], name):
         """
         register a configuration class
-        :param cls:
+        :param config:
         :param name:
         :return:
         """
-        self._configs.add(cls)
+        self._configs.add(config)
         self._config_names.add(name)
+        # update the attributes_to_config map
+        for attribute in config.get_attributes():
+            if attribute in self.attribute_to_config:
+                raise ValueError("pytconf: attribute [{}] appears more than once".format(attribute))
+            self.attribute_to_config[attribute] = config
 
     @classmethod
-    def print_errors(cls, errors: List[str]) -> None:
-        if errors:
-            print()
-            for error in errors:
-                print_warn(error)
-            print()
+    def print_errors(cls, errors: ErrorsCollector) -> None:
+        for error in errors.errors:
+            print_error(error)
 
     def show_help(self) -> None:
         print("Usage: {} [OPTIONS] COMMAND [ARGS]...".format(self.main_function.__name__))
@@ -94,7 +153,7 @@ class PytconfConf(object):
 
     def show_help_for_function(self, function_name: str, show_help_full: bool, show_help_suggest: bool) -> None:
         print("Usage: {} {} [OPTIONS] [ARGS]...".format(
-            self.main_function.__name__,
+            self.app_name,
             function_name,
         ))
         function_selected = self.function_name_to_callable[function_name]
@@ -141,76 +200,75 @@ class PytconfConf(object):
                 print("      {}".format(more_help))
         print()
 
-    def parse_args(self, command_selected: str, flags: Dict[str, str], _free_args: List[str]) -> bool:
+    def parse_args(
+        self,
+        command_selected: str,
+        flags: Dict[str, str],
+        errors: ErrorsCollector,
+    ) -> None:
         """
         Parse the args and fill the global data
-        Currently we disregard the free parameters
         :param command_selected:
         :param flags:
-        :param _free_args:
-        :return:
+        :param errors:
         """
-        configs = self.function_name_to_configs[command_selected]
-        suggested_configs = self.function_name_to_suggest_configs[command_selected]
 
-        # create the attribute_to_config map
-        attribute_to_config: Dict[str, Config] = dict()
-        for config in itertools.chain(configs, suggested_configs):
-            for attribute in config.get_attributes():
-                if attribute in attribute_to_config:
-                    raise ValueError("attribute [{}] double".format(attribute))
-                attribute_to_config[attribute] = config
-
-        # set the flags into the "default" field
+        # set the flags into the "default" field and collect unknown flags
         unknown_flags = []
         for flag_raw, value in flags.items():
-            edit = value.startswith('=')
-            if flag_raw not in attribute_to_config:
+            if flag_raw not in self.attribute_to_config:
                 unknown_flags.append(flag_raw)
-            config = attribute_to_config[flag_raw]
+                continue
+            config = self.attribute_to_config[flag_raw]
             param = config.get_param_by_name(flag_raw)
+            edit = value.startswith('=')
             if edit:
                 v = param.s2t_generate_from_default(value[1:])
             else:
                 v = param.s2t(value)
             setattr(config, flag_raw, v)
+        if unknown_flags:
+            errors.add_error("unknown flags [{}]".format(",".join(unknown_flags)))
 
-        # check for missing parameters and show help if there are any missing
+        # check for missing parameters
         missing_parameters = []
+        configs = self.function_name_to_configs[command_selected]
         for config in configs:
             for attribute in config.get_attributes():
                 value = getattr(config, attribute)
                 if value is NO_DEFAULT:
                     missing_parameters.append(attribute)
-        if unknown_flags or missing_parameters:
-            if missing_parameters:
-                print()
-                print_warn("missing parameters [{}]".format(",".join(missing_parameters)))
-            if unknown_flags:
-                print()
-                print_warn("unknown flags [{}]".format(",".join(unknown_flags)))
-            print("problems found, not running")
-            print()
-            self.show_help_for_function(command_selected, show_help_full=False, show_help_suggest=False)
-            return False
+        if missing_parameters:
+            errors.add_error("missing parameters [{}]".format(",".join(missing_parameters)))
 
-        # move all default values to place
+        # move all default values to place (this will not be needed in the new scheme)
         for config in itertools.chain(configs, self._configs):
             for attribute in config.get_attributes():
                 param: Param = getattr(config, attribute)
                 if isinstance(param, Param):
                     if param.default is not NO_DEFAULT:
                         setattr(config, attribute, param.default)
-        return True
 
-    def config_arg_parse_and_launch(self, print_messages=True, launch=True) -> None:
+    def config_arg_parse_and_launch(
+        self,
+        args: Union[List[str], None] = None,
+        launch=True
+    ) -> None:
+        # if we are given no args then take the from sys.argv
+        if args is None:
+            args = sys.argv
+
         # we don't need the first argument which is the script path
-        args: List[str] = sys.argv[1:]
+        if args:
+            self.app_name = args[0]
+            args = args[1:]
+        else:
+            self.app_name = "UNKNOWN APP NAME"
         # name of arg and it's value
         flags: Dict[str, str] = dict()
         special_flags = set()
-        errors: List[str] = []
-        free_args = []
+        errors = ErrorsCollector()
+        self.free_args = []
         while args:
             current = args.pop(0)
             if current.startswith("--"):
@@ -226,18 +284,15 @@ class PytconfConf(object):
                     elif real in SPECIAL_COMMANDS:
                         special_flags.add(real)
                     else:
-                        errors.append("argument [{}] needs a follow-up argument".format(real))
+                        errors.add_error("argument [{}] needs a follow-up argument".format(real))
                 else:
-                    errors.append("can not parse argument [{}]".format(real))
+                    errors.add_error("can not parse argument [{}]".format(real))
             else:
-                free_args.append(current)
+                self.free_args.append(current)
         show_help = False
         show_help_full = False
         show_help_suggest = False
 
-        if len(errors) > 0:
-            show_help = True
-
         if "help" in special_flags:
             show_help = True
         if "help-suggest" in special_flags:
@@ -247,34 +302,46 @@ class PytconfConf(object):
             show_help = True
             show_help_full = True
 
+        # find the selected command
         command_selected = None
-        if len(free_args) >= 1:
-            command = free_args.pop(0)
+
+        # if there are free args then the command is the first of these
+        if len(self.free_args) >= 1:
+            command = self.free_args.pop(0)
             if command in self.function_name_to_callable:
                 command_selected = command
             else:
-                errors.append("Unknown command [{}]".format(command))
+                errors.add_error("Unknown command [{}]".format(command))
 
+        # if there are no free args and just one function then this is it
         if len(self.function_name_to_callable) == 1:
-            for name in self.function_name_to_callable.keys():
-                command_selected = name
+            command_selected = list(self.function_name_to_callable.keys())[0]
 
-        if len(free_args) > 0:
-            errors.append("free args are not allowed")
+        # check if we are not allowed free args
+        if command_selected is not None:
+            if not self.allow_free_args[command_selected] and len(self.free_args) > 0:
+                errors.add_error("free args are not allowed")
 
-        if print_messages:
-            if show_help or errors or command_selected is None:
-                self.print_errors(errors)
-                if command_selected:
-                    self.show_help_for_function(command_selected, show_help_full, show_help_suggest)
-                else:
-                    self.show_help()
+        if command_selected is None:
+            errors.add_error("no command is selected")
+        else:
+            self.parse_args(command_selected, flags, errors)
+
+        if errors.have_errors():
+            self.print_errors(errors)
+            if not show_help:
                 return
 
-        f = self.function_name_to_callable[command_selected]
-        if self.parse_args(command_selected, flags, free_args):
-            if launch:
-                f()
+        if show_help:
+            if command_selected:
+                self.show_help_for_function(command_selected, show_help_full, show_help_suggest)
+            else:
+                self.show_help()
+            return
+
+        if launch:
+            function_to_run = self.function_name_to_callable[command_selected]
+            function_to_run()
 
     def get_html(self) -> str:
         html_gen = HtmlGen()
@@ -342,67 +409,16 @@ def get_pytconf():
     return _pytconf
 
 
-def config_arg_parse_and_launch(print_messages=True, launch=True):
+def config_arg_parse_and_launch(launch=True, args=None) -> None:
     """
-    backwards compatibility function
-    :return:
+    This is the real API
     """
     get_pytconf().config_arg_parse_and_launch(
-        print_messages=print_messages,
         launch=launch,
+        args=args,
     )
 
 
-class MetaConfig(type):
-    """
-    base class for all configs
-    """
-
-    def __new__(mcs, name, bases, cls_dict):
-        if name != "Config":
-            params_dict = dict()
-            cls_dict[PARAMS_ATTRIBUTE] = params_dict
-            for k, d in cls_dict.items():
-                if isinstance(d, Param):
-                    # assert d.default is not NO_DEFAULT
-                    params_dict[k] = d
-                    cls_dict[k] = d.default
-        return type.__new__(mcs, name, bases, cls_dict)
-
-    def __init__(cls, name, bases, cls_dict):
-        # print(name, bases, cls_dict)
-        # if len(cls.mro()) > 2:
-        #     register_config(cls, name)
-        #     # print("was subclassed by " + name)
-        if name != "Config":
-            get_pytconf().register_config(cls, name)
-        # print(name, cls_dict)
-        super(MetaConfig, cls).__init__(name, bases, cls_dict)
-
-
-ParamType = TypeVar('ParamType', bound='Param')
-
-
-class Config(metaclass=MetaConfig):
-    """
-        base class for all configs
-    """
-
-    @classmethod
-    def get_attributes(cls: Any) -> List[str]:
-        return getattr(cls, PARAMS_ATTRIBUTE).keys()
-        # return [attr for attr in dir(cls) if not callable(getattr(cls, attr))
-        #        and not attr.startswith("__")]
-
-    @classmethod
-    def get_params(cls: Any) -> Dict[str, ParamType]:
-        return getattr(cls, PARAMS_ATTRIBUTE)
-
-    @classmethod
-    def get_param_by_name(cls: Any, name: str) -> ParamType:
-        return cls.get_params()[name]
-
-
 class Unique:
     pass
 
@@ -613,9 +629,9 @@ class ParamCreator(object):
             help_string=help_string,
             default=default,
             type_name="int",
-            function_s2t=convert_string_to_int,
-            function_t2s=convert_int_to_string,
-            function_s2t_generate_from_default=convert_string_to_int_default,
+            function_s2t=convert_str_to_int,
+            function_t2s=convert_int_to_str,
+            function_s2t_generate_from_default=convert_str_to_int_default,
         )
 
     @staticmethod
@@ -634,8 +650,8 @@ class ParamCreator(object):
             help_string=help_string,
             default=default,
             type_name="List[int]",
-            function_s2t=convert_string_to_list_int,
-            function_t2s=convert_list_int_to_string,
+            function_s2t=convert_str_to_list_int,
+            function_t2s=convert_list_int_to_str,
         )
 
     @staticmethod
@@ -654,8 +670,8 @@ class ParamCreator(object):
             help_string=help_string,
             default=default,
             type_name="List[str]",
-            function_s2t=convert_string_to_list_str,
-            function_t2s=convert_list_str_to_string,
+            function_s2t=convert_str_to_list_str,
+            function_t2s=convert_list_str_to_str,
         )
 
     @staticmethod
@@ -674,8 +690,8 @@ class ParamCreator(object):
             help_string=help_string,
             default=default,
             type_name="Union[int, None]",
-            function_s2t=convert_string_to_int_or_none,
-            function_t2s=convert_int_or_none_to_string,
+            function_s2t=convert_str_to_int_or_none,
+            function_t2s=convert_int_or_none_to_str,
         )
 
     @staticmethod
@@ -694,8 +710,28 @@ class ParamCreator(object):
             help_string=help_string,
             default=default,
             type_name="str",
-            function_s2t=convert_string_to_string,
-            function_t2s=convert_string_to_string,
+            function_s2t=convert_str_to_str_or_none,
+            function_t2s=convert_str_or_none_to_str,
+        )
+
+    @staticmethod
+    def create_str_or_none(
+        help_string: str = NO_HELP,
+        default: Union[str, None, NO_DEFAULT_TYPE] = NO_DEFAULT
+    ) -> Union[str, None]:
+        """
+        Create a string parameter
+        :param help_string:
+        :param default:
+        :return:
+        """
+        # noinspection PyTypeChecker
+        return ParamFunctions(
+            help_string=help_string,
+            default=default,
+            type_name="str",
+            function_s2t=convert_str_to_str,
+            function_t2s=convert_str_to_str,
         )
 
     @staticmethod
@@ -714,8 +750,8 @@ class ParamCreator(object):
             help_string=help_string,
             default=default,
             type_name="bool",
-            function_s2t=convert_string_to_bool,
-            function_t2s=convert_bool_to_string,
+            function_s2t=convert_str_to_bool,
+            function_t2s=convert_bool_to_str,
         )
 
     @staticmethod
@@ -854,8 +890,8 @@ class ParamCreator(object):
             help_string=help_string,
             default=default,
             type_name="bucket_name",
-            function_s2t=convert_string_to_string,
-            function_t2s=convert_string_to_string,
+            function_s2t=convert_str_to_str,
+            function_t2s=convert_str_to_str,
         )
 
 
@@ -874,17 +910,42 @@ def register_main() -> Callable[[Any], Any]:
 
 
 def register_endpoint(
-    configs: List[Config] = (),
-    suggest_configs: List[Config] = (),
-    group: str = DEFAULT_GROUP_NAME
+    configs: List[Type[Config]] = (),
+    suggest_configs: List[Type[Config]] = (),
+    group: str = DEFAULT_GROUP_NAME,
+    allow_free_args: bool = False,
 ) -> Callable[[Any], Any]:
+    logger = logging.getLogger("pytconf")
+    logger.debug("registering endpoint")
+
     def identity(f):
-        pt = get_pytconf()
-        function_name = f.__name__
-        pt.function_name_to_callable[function_name] = f
-        pt.function_name_to_configs[function_name] = configs
-        pt.function_name_to_suggest_configs[function_name] = suggest_configs
-        pt.function_group_names[group].add(function_name)
+        register_function(
+            f,
+            configs=configs,
+            suggest_configs=suggest_configs,
+            group=group,
+            allow_free_args=allow_free_args,
+        )
         return f
 
     return identity
+
+
+def register_function(
+    f: Callable,
+    configs: List[Type[Config]] = (),
+    suggest_configs: List[Type[Config]] = (),
+    group: str = DEFAULT_GROUP_NAME,
+    allow_free_args: bool = False,
+) -> None:
+    function_name = f.__name__
+    pt = get_pytconf()
+    pt.function_name_to_callable[function_name] = f
+    pt.function_name_to_configs[function_name] = configs
+    pt.function_name_to_suggest_configs[function_name] = suggest_configs
+    pt.function_group_names[group].add(function_name)
+    pt.allow_free_args[function_name] = allow_free_args
+
+
+def get_free_args() -> List[str]:
+    return get_pytconf().free_args
